import dgl
import torch as th
import torch.nn.functional as F
from datasets import get_dataset
from model import GAT, GraphSage, GCN,JKNet
import time
from torchmetrics.functional import accuracy
import numpy as np
from sampler import sampler as sampler_
import sys
import argparse
from args import get_args,get_args2,get_args3,get_args4
import dgl.data as data


def eval(model, graph):
    with th.no_grad():
        model.eval()
        mask = g_all.ndata['test_mask']
        labels = g_all.ndata['label']
        logits = model(graph)
        logits = logits[mask]
        labels = labels[mask]
        return accuracy(logits, labels).item()

parser = argparse.ArgumentParser('main')
parser.add_argument('--sample-type', type=str, default = 'uniform_sample',
                    help='the type of sample')
parser.add_argument('--p1', type=float,help='the value of parameter p1')
parser.add_argument('--p2', type=float,help='the value of parameter p2')
parser.add_argument('--q', type=float,help='the value of parameter q')
parser.add_argument('--lambda_', type=float,help='the value of parameter lambda')
parser.add_argument('--uniform-ratio',default='1.0', type=float,help='the ratio of uniform sampling')
parser.add_argument('--model-name', type=str,
                    choices=['gcn', 'gat', 'graphsage','jknet'],
                    help='the name of model')
parser.add_argument("--gpu", type=int, default=-1,
                    help="which GPU to use. Set -1 to use CPU.")
parser.add_argument('--dataset', type=str, default='cora_full',
                    choices=['cora', 'citeseer', 'pubmed','reddit'],
                    help='the name of dataset')
parser.add_argument("--out-head", type=int, default=1,
                    help="number of output attention heads")
parser.add_argument("--heads", type=str, default='8',
                    help="list of the numbers of heads splited by '-' ")
parser.add_argument("--hidden-dims", type=str, default='8',
                    help="list of the numbers of heads splited by '-' ")
parser.add_argument("--residual", action="store_true", default=False,
                    help="use residual connection")
parser.add_argument("--bias", action="store_true", default=False,
                    help="use bias")
parser.add_argument("--feat-drop", type=float, default=.0,
                    help="input feature dropout")
parser.add_argument("--attn-drop", type=float, default=.0,
                    help="attention dropout")
parser.add_argument('--negative-slope', type=float, default=0.2,
                    help="the negative slope of leaky relu")
parser.add_argument("--epochs", type=int, default=100,
                    help="number of training epochs")
parser.add_argument("--lr", type=float, default=0.005,
                    help="learning rate")
parser.add_argument('--weight-decay', type=float, default=5e-4,
                    help="weight decay")
args_model = parser.parse_args()
if args_model.model_name == 'gat':
    args = get_args()
elif args_model.model_name == 'graphsage':
    args = get_args2()
elif args_model.model_name == 'gcn':
    args = get_args3()
elif args_model.model_name == 'jknet':
    args = get_args4()
if args.gpu > 0:
    device = 'cuda:0'
else:
    device = 'cpu'
print('dataset: {}'.format(args.dataset))

g, num_feats, n_classes = get_dataset(args.dataset, device)
g = dgl.remove_self_loop(g)
g_all = dgl.add_self_loop(g)



if args_model.model_name == 'gat':
    hidden_dims = args.hidden_dims.split('-')
    hidden_dims = [int(hidden_dim) for hidden_dim in hidden_dims]
    heads = args.heads.split('-')
    heads = [int(head) for head in heads]

    model = GAT(num_feats,
                hidden_dims,
                n_classes,
                heads,
                args.out_head,
                F.elu,
                args.feat_drop,
                args.attn_drop,
                args.negative_slope,
                args.residual,
                args.bias)
elif args_model.model_name == 'graphsage':
    hidden_dims = args.hidden_dims.split('-')
    hidden_dims = [int(hidden_dim) for hidden_dim in hidden_dims]
    model = GraphSage(num_feats,
                      hidden_dims,
                      n_classes,
                      dropout=args.feat_drop)
elif args_model.model_name == 'gcn':
    hidden_dims = args.hidden_dims.split('-')
    hidden_dims = [int(hidden_dim) for hidden_dim in hidden_dims]
    model = GCN(num_feats,
                hidden_dims,
                n_classes,
                dropout=args.feat_drop)
elif args_model.model_name == 'jknet':
    model = JKNet(in_dim=num_feats,
                  hid_dim=args.hid_dim,
                  out_dim=n_classes,
                  num_layers=args.num_layers,
                  mode=args.mode,
                  dropout=args.dropout).to(device)
print(model)
model.to(device)

loss_fcn = th.nn.CrossEntropyLoss()
if args_model.model_name in ['gat','gcn','graphsage']:
    optimizer = th.optim.Adam(
        model.parameters(), lr=args.lr, weight_decay=args.weight_decay)
elif  args_model.model_name  == 'jknet':
    optimizer = th.optim.Adam(model.parameters(), lr=args.lr, weight_decay=args.lamb)

labels = g.ndata['label']
time_list = []


sample_type = args.sample_type
# uniform_sample    gds
sampler = sampler_(g)

uniform_sample_ratio = args.uniform_ratio

p1 = args.p1
p2 = args.p2
q = args.q
print('model:{},sample_type:{}, uniform_sample_ratio:{}, para_p:{}, para_q:{}, quantile:{}'.format(
    args_model.model_name, sample_type, uniform_sample_ratio, p1, p2, q))



if sample_type == 'uniform_sample':
    subgraph, ratio = sampler.uniform_sample(uniform_sample_ratio)

elif sample_type == 'gds':
    subgraph, ratio = sampler.gds(p1=p1, p2=p2, q=q)

subgraph = dgl.add_self_loop(subgraph)
labels = subgraph.ndata['label'].to(device)
train_mask = subgraph.ndata['train_mask'].to(device)
val_mask = subgraph.ndata['val_mask'].to(device)
if args_model.model_name in ['gat', 'gcn', 'graphsage']:
    model.reset_parameters()
else:
    model.reset_params()

val_acc_list = []
acc_list = []
for epoch in range(args.epochs):
    start = time.time()

    model.train()
    subgraph = subgraph.to(device)
    logits = model(subgraph)
    loss = loss_fcn(logits[train_mask], labels[train_mask])
    optimizer.zero_grad()
    loss.backward()
    optimizer.step()

    train_acc = accuracy(logits[train_mask], labels[train_mask])
    time_list.append(time.time() - start)

    val_acc = accuracy(logits[val_mask], labels[val_mask])
    val_acc_list.append(val_acc)

    print("Epoch {:05d} | Time(s) {:.4f} | Loss {:.4f} | TrainAcc {:.4f} |"
          " ValAcc {:.4f} | ratio {:.4f}".
          format(epoch, np.mean(time_list), loss.item(), train_acc,
                  val_acc, ratio))
    g_all = g_all.to(device)
    acc = eval(model, g_all)
    g_all = g_all.to('cpu')
    acc_list.append(acc)
val_acc_list = np.array(val_acc_list)
result_acc = acc_list[val_acc_list.argmax()]
result_acc_list.append(result_acc)
print('test acc:{} , ratio: {}'.format(result_acc, ratio))


